import logging

from service_manager.service_initializer.creators.bamboo_build_creator import BambooBuildCreator
from service_manager.service_initializer.creators.cookie_cutter_creator import CookeCutterProjectCreator
from service_manager.util import services
from service_manager.util.services import walk_service_map, ensure_service_directory_exists, Service


class Initializer(object):
    def __init__(self, vcs, destination_directory, dry_run,service_template_definitions):
        super(Initializer, self).__init__()
        self.project_creator = CookeCutterProjectCreator(service_template_definitions, dry_run)
        self.build_creator = BambooBuildCreator(service_template_definitions, dry_run)
        self.dry_run = dry_run
        services.safe_mkdir(destination_directory=destination_directory)
        self.destination_directory = destination_directory
        self.vcs = vcs

    def init_app(self,app):
        pass
    
    def init_service(self, definition):
        # type: (Service ) -> None
        if definition.repo_exists():
            logging.info("Service exists - {}".format(definition.get_fully_qualified_service_name()))
            if definition.force_recreate_build():
                self.build_creator.create_project(definition)
            return
        logging.info("Creating Service -{}".format(definition.get_fully_qualified_service_name()))
        destination_dir = ensure_service_directory_exists(self.destination_directory, service_defintion=definition)
        services.pretty_print_service(definition)
        self.project_creator.create_project(definition,destination_dir)
        self.vcs.create_project(definition,destination_dir)
        self.build_creator.create_project(definition)

    def initialize_services(self, application_map):
        walk_service_map(application_map=application_map, application_callback=self.init_app,
                         service_callback=self.init_service)



import json
import os

import logging
from cookiecutter.main import cookiecutter


def _make_cookie_safe(service_definition):
    ret = {}
    for key,valey in service_definition.iteritems():
        ret[key.replace('-',"_")] = valey
    return ret


class CookeCutterProjectCreator(object):
    def __init__(self, template_dir,dry_run):
        super(CookeCutterProjectCreator, self).__init__()
        self.dry_run = dry_run
        self.service_def = {}
        self._load_service_templates(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), 'builtin_service_templates.json'))
        if template_dir:
            self.template_dir = template_dir
            self._load_service_templates(os.path.join(template_dir,'template-definitions.json'))

    def _load_service_templates(self, builtIn):
        with open(builtIn) as builtin:
            self.service_def.update(json.load(builtin))

    def create_project(self, service_definition, service_dir):
        template = self.lookup_service_template(service_definition.get_service_type())
        if template['type'] == 'file':
            location = os.path.abspath(os.path.join(self.template_dir,template['location']))
        else:
            location = template['location']
        extra_context = _make_cookie_safe(service_definition)
        if self.dry_run:
            logging.error("Creating project from template {} ".format(location))
        else:
            return cookiecutter(location,
                                no_input=True,
                                extra_context=extra_context,
                                output_dir=service_dir)

    def lookup_service_template(self, service_type):
        if service_type not in self.service_def:
            raise Exception("Unknown service type - {}".format(service_type))
        return self.service_def[service_type]



import json
import logging
import os
import subprocess
from collections import defaultdict

from service_manager.util.log_handler import print_red_bold

REPO_URL = 'repository_url'

REPOSITORY_NAME = 'repository'
REPOSITORY_URL = 'repository_url'
DESCRIPTION = 'description'
ROLE = 'role'
APPLICATION = 'application'
FQN = 'project_name'
FORCE_RECREATE_BUILD = 'recreate_build'
SERVICE_TYPE = 'service-type'


class Application(dict):
    def __init__(self, application):
        super(Application, self).__init__()
        self.application = application

    def add_service(self, role, service):
        self[role] = service

    def get_contract_test_git_url(self):
        for name,definition in self.iteritems():
            if definition.get_service_type() == 'contract-tests':
                return definition.get_git_url()


class Service(dict):
    def __init__(self, app, role, definition, app_reference=None):
        super(Service, self).__init__()
        self.update(definition)
        self.app_ref = app_reference
        self[APPLICATION] = app
        self[ROLE] = role
        self[FQN] = self.get_fully_qualified_service_name()

    def get_fully_qualified_service_name(self):
        return "{application}-{role}".format(**self).replace(' ','_')

    def force_recreate_build(self):
        return self.get(FORCE_RECREATE_BUILD,False)

    def get_description(self):
        return self[DESCRIPTION]

    def get_service_type(self):
        return self[SERVICE_TYPE]

    def repo_exists(self):
        return REPOSITORY_URL in self

    def get_role(self):
        return self[ROLE]

    def get_app(self):
        return self[APPLICATION]

    def get_repository_name(self):
        return self[REPOSITORY_NAME] if REPOSITORY_NAME in self else self.get_fully_qualified_service_name()

    def set_git_url(self, url):
        self[REPO_URL] = url

    def get_git_url(self):
        return self.get(REPO_URL,None)

    def get_contract_test_git_url(self):
        return self.app_ref.get_contract_test_git_url()

    def get_service_directory(self,app_dir):
        return os.path.join(app_dir,self.get_fully_qualified_service_name())


def load_service_definitions(service_directory, app_filter=None):
    service_dir = os.path.abspath(service_directory)
    listdir = os.listdir(service_dir)
    service_map = defaultdict(dict)
    for dir in listdir:
        if _is_valid_app(dir, service_dir) and application_filter(app_filter, dir):
            service_definition_file = os.path.join(service_dir, dir, 'service.json')
            if not os.path.exists(service_definition_file):
                logging.warn("Found invalid application directory - no service.json exists - {}".format(service_dir))
            else:
                service_map[dir] = Application(dir)
                with open(service_definition_file) as service_def:
                    service_definitions = json.load(service_def)
                    for role, definition in service_definitions.iteritems():
                        service_map[dir].add_service(role, Service(app=dir, role=role, definition=definition, app_reference=service_map[dir]))
    return service_map


def _is_valid_app(dir, service_dir):
    return os.path.isdir(os.path.join(service_dir, dir)) and not dir.startswith('.')


def application_filter(app_filter, application):
    return app_filter is None or application.startswith(app_filter)


def walk_service_map(application_map, application_callback, service_callback):
    for application, service_map in application_map.iteritems():
        if application_callback: application_callback(application)
        for service, service_definition in service_map.iteritems():
            if service_callback: service_callback(service_definition)


def pretty_print_application(app): logging.error(u"{}".format(app))


def pretty_print_service(service_definition):
    logging.warn(u"\t {}".format(service_definition.get_role()))
    for dat in service_definition:
        secondary_indent = '\t\t' if len(dat) <= 10 else '\t'
        logging.info(u"\t\t {}{}- {}".format(dat, secondary_indent, service_definition[dat]))


def pretty_print_services(application_map):
    walk_service_map(application_map, pretty_print_application, pretty_print_service)


def safe_mkdir(destination_directory):
    if not os.path.exists(destination_directory):
        os.mkdir(destination_directory)


def ensure_service_directory_exists(destination_directory, service_defintion):
    app_dir = os.path.join(destination_directory, service_defintion.get_app())
    safe_mkdir(app_dir)
    return app_dir


def invoke_process(args, exec_dir=None, dry_run=False):
    if dry_run:
        print_red_bold(u"\t {}".format(str(args)))
        return 0
    else:
        arg_list = {'args': args}
        if exec_dir:
            arg_list['cwd'] = exec_dir
        return subprocess.call(**arg_list)



import argparse
import json
import os

from service_initializer.creators.cookie_cutter_creator import CookeCutterProjectCreator
from service_initializer.initializer import Initializer
from util.log_handler import configure_logging
from util.services import load_service_definitions, pretty_print_services
from vcs.vcs import VCS

LIST = 'list'
SYNC = 'sync'
PULL = 'pull'
parser = argparse.ArgumentParser(prog="service-manager", description='Utility to help deploy and manage microservices.')

parser.add_argument('--verbose', help='Level of logging to output', action="store_true", default=False)
parser.add_argument('--application-filter', help='Constrain to passed application', type=str, default=None)
parser.add_argument('--service-directory', help='Directory containing service definitions in <app>/service.json format',
                    type=str, required=True)
parser.add_argument('--dry-run', help='Preview effect of action', action="store_true",
                    default=False)
subparsers = parser.add_subparsers(help='commands')

list_parser = subparsers.add_parser(LIST, help='List known services')
list_parser.set_defaults(command=LIST)
list_parser.add_argument('--validate-repository', help='Validate existence of repository in VCS', action="store_true",
                         default=False)
sync_parser = subparsers.add_parser(SYNC, help='Sync service definitions and initialize new entries with templated code'
                                               ' in VCS and Build System.')

sync_parser.set_defaults(command=SYNC)
sync_parser.add_argument('--destination-directory', help='Destination directory to create new service definition.',
                         required=True)
sync_parser.add_argument('--service-template-definitions',
                         help='File containing references to custom service templates.'
                              '  JSON dictionary {service-type: {type: file|git, location: relative path | git URL}}',
                         type=str, default=None)

pull_parser = subparsers.add_parser(PULL, help='Pull git repos into local filesystem.')
pull_parser.set_defaults(command=PULL)
pull_parser.add_argument('--destination-directory', help='Destination directory to pull code into.', default=".")



def execute_main():
    args = parser.parse_args()
    configure_logging(args.verbose)
    application_map = load_service_definitions(args.service_directory, args.application_filter)
    vcs = VCS(args.service_directory, args.dry_run)
    if args.command == LIST:
        if args.validate_repository:
            vcs.validate_repositories(application_map)
        pretty_print_services(application_map)
    elif args.command == SYNC:
        vcs.validate_repositories(application_map)
        init = Initializer(vcs, args.destination_directory, args.dry_run,args.service_template_definitions)
        init.initialize_services(application_map)
    elif args.command == PULL:
        vcs.validate_repositories(application_map)
        vcs.pull_services(application_map, args.destination_directory)


if __name__ == "__main__":
    execute_main()



import logging


def configure_logging(verbose=False):
    console = ColorLog()
    console.setLevel(logging.DEBUG if verbose else logging.INFO)
    logging.getLogger('').addHandler(console)
    logging.getLogger('').setLevel(logging.DEBUG)


def _wrap_with(code):
    def inner(text, bold=False):
        c = code
        if bold:
            c = "1;%s" % c
        return u"\033[%sm%s\033[0m" % (c, text)

    return inner


red = _wrap_with('31')
green = _wrap_with('32')

yellow = _wrap_with('33')
blue = _wrap_with('34')
magenta = _wrap_with('35')
cyan = _wrap_with('36')
white = _wrap_with('37')


def print_color(color,bold=False):
    def inner(text):
        print(color(text,bold))
    return inner


print_red = print_color(red)
print_red_bold = print_color(red,True)


class ColorLog(logging.Handler):
    """
    A class to print colored messages to stdout
    """
    COLORS = {
        logging.CRITICAL: red,
        logging.ERROR: red,
        logging.WARNING: yellow,
        logging.INFO: green,
        logging.DEBUG: lambda x: x,
    }

    def __init__(self):
        logging.Handler.__init__(self)

    def usesTime(self):
        return False

    def emit(self, record):
        color = self.COLORS.get(record.levelno, lambda x: x)
        print(color(record.msg))
